#ifdef WIN32
#include <windows.h>
#endif //WIN32
#include <string.h>
#include <malloc.h>
#include <stdio.h>
#include <conio.h>
#include "pcre.h"

#define MAX_URL_LENGTH  32768
#define REGEXP_N_GRAN   32
#define MAX_MATCH_ELEMENTS  128

char* url_buffer;
char* url2_buffer;
char* in_name;
char* log_name;
int mthread=0;

void usage() {
    printf(
           "Squid URL Rewrite v1.0 (c) by Alexander A. Telyatnikov (Alter), 2012-2013\n"
           "  Home site http://alter.org.ua\n"
           "Usage:\n"
           "  url_rewrite [-<switches>] [-f|conf=]<config file>\n"
           "Switches:\n"
/*           "  -m                  multi-thread mode\n"*/
           "  -l <logfile>        rewrite log file\n"
           "  -?,-h               display this help\n"
          );
    exit(0);
}

void do_log(char* str, int len) {
    FILE* lfh;

    if(!log_name) {
        return;
    }
    lfh = fopen(log_name, "a+");
#ifdef WIN32
    if(!lfh) {
        lfh = fopen(log_name, "w+");
    }
#endif //WIN32
    if(!lfh) {
        return;
    }
    fwrite(str, len, 1, lfh);
    fclose(lfh);
}

#ifdef WIN32
HANDLE StdIn;

int getch() {
    UCHAR buff[2];
    ULONG read_bytes;

    if(!ReadFile(StdIn, &buff[0], 1, &read_bytes, NULL)) {
        return -1;
    }
    return buff[0];
}
#endif //WIN32

int main(int argc, char** argv) {
    int i, j;

    unsigned int len, plen, ilen;
    char* pos;
    int break_sw=0;
    char a;

    pcre* pcre_regexp1;
    const char *error;
    int erroffset;

    pcre** pcre_regexp_arr;
    pcre** pcre_regexp_arr_tmp;

    char** pattern_arr;
    char** pattern_arr_tmp;

    char*  pattern_buff;
    char*  pattern_buff_tmp;
    int    pattern_buff_max_sz;
    int    pattern_buff_sz;

    int max_regexp;
    int regexp_count;

    char* regexp_start;
    char* regexp_end;
    char* pattern_start;
    char* pattern_end;

    int ovector[3*MAX_MATCH_ELEMENTS];
    int ovector_max_sz = 3*MAX_MATCH_ELEMENTS;

    int rc;
    char* url2_buffer_ptr;

#ifdef WIN32
    ULONG InputMode;

    StdIn  = GetStdHandle(STD_INPUT_HANDLE);
    if(GetConsoleMode(StdIn, &InputMode)) {
        if(!SetConsoleMode(StdIn, InputMode & ~(ENABLE_LINE_INPUT | ENABLE_ECHO_INPUT))) {
            printf("Can't set console input mode\n");
            return 0;
        }
    }
#endif //WIN32

    in_name = NULL;
    log_name = NULL;
    FILE* fh;

    url_buffer = (char*)malloc(MAX_URL_LENGTH);
    if(!url_buffer) {
        printf("Can't allocate memory\n");
        return 0;
    }
    url2_buffer = (char*)malloc(MAX_URL_LENGTH);
    if(!url2_buffer) {
        printf("Can't allocate memory\n");
        return 0;
    }

    if(argc < 2)
        usage();

    for(i=1; i<argc; i++) {
        if(!argv[i])
            continue;
        if(argv[i][0] != '-') {
            if(!strncmp(argv[i], "conf=", 5)) {
                in_name = argv[i]+5;
                continue;
            } else
            if(!strncmp(argv[i], "log=", 4)) {
                log_name = argv[i]+4;
                continue;
            } else
            if(!in_name) {
                in_name = argv[i];
                continue;
            } else {
                usage();
            }
        }
        j=1;
        break_sw = 0;
        while(!break_sw &&
              (a = argv[i][j]) && (a != ' ') && (a != '\t')) {
            switch(a) {
            case 'f' :
                i++;
                if(!argv[i])
                    usage();
                in_name = argv[i];
                break_sw = 1;
                break;
            case 'l' :
                i++;
                if(!argv[i])
                    usage();
                log_name = argv[i];
                break_sw = 1;
                break;
            case 'm' :
                mthread = 1;
                break_sw = 1;
                break;
            case '?' :
            default:
                usage();
            }
            j++;
        }
    }

    if(!in_name)
        usage();

    fh = fopen(in_name, "r");
    if(!fh) {
        printf("Can't open %s\n", in_name);
        return 0;
    }

    max_regexp=REGEXP_N_GRAN;
    regexp_count=0;
    pcre_regexp_arr = (pcre**)malloc(max_regexp*sizeof(pcre*));
    if(!pcre_regexp_arr) {
        printf("Can't allocate memory\n");
        return 0;
    }

    pattern_arr = (char**)malloc(max_regexp*sizeof(char*));
    if(!pattern_arr) {
        printf("Can't allocate memory\n");
        return 0;
    }

    pattern_buff_max_sz = MAX_URL_LENGTH*2;
    pattern_buff_sz = 0;
    pattern_buff = (char*)malloc(pattern_buff_max_sz);
    if(!pattern_buff) {
        printf("Can't allocate memory\n");
        return 0;
    }

    // read config
    while(fgets(url_buffer, MAX_URL_LENGTH-1, fh)) {
        url_buffer[MAX_URL_LENGTH-1]=0;
        if(regexp_count >= max_regexp) {
            max_regexp += REGEXP_N_GRAN;

            // reallocate regexp pointers
            pcre_regexp_arr_tmp = (pcre**)malloc(max_regexp*sizeof(pcre*));
            if(!pcre_regexp_arr_tmp) {
                printf("Can't reallocate memory\n");
                return 0;
            }
            memcpy(pcre_regexp_arr_tmp, pcre_regexp_arr, regexp_count*sizeof(pcre*));
            free(pcre_regexp_arr);
            pcre_regexp_arr = pcre_regexp_arr_tmp;

            // reallocate pattern pointers
            pattern_arr_tmp = (char**)malloc(max_regexp*sizeof(char*));
            if(!pattern_arr_tmp) {
                printf("Can't reallocate memory\n");
                return 0;
            }
            memcpy(pattern_arr_tmp, pattern_arr, regexp_count*sizeof(char*));
            free(pattern_arr_tmp);
            pattern_arr = pattern_arr_tmp;
        }

        regexp_start = strchr((const char*)url_buffer, '@');
        if(!regexp_start) {
            printf("No regexp start marker, skipping %s\n", url_buffer);
            continue;
        }
        regexp_start++;
        regexp_end = strchr((const char*)regexp_start, '@');
regexp_end_retry:
        if(!regexp_end) {
            printf("No regexp end marker, skipping %s\n", url_buffer);
            continue;
        }
        if(regexp_end[-1] == '\\') {
            regexp_end = strchr((const char*)regexp_end+1, '@');
            goto regexp_end_retry;
        }

        regexp_end[0] = 0; // terminator for regexp
        pattern_start = regexp_end+1;
        pattern_end = strchr((const char*)pattern_start, '@');
        if(!pattern_end) {
            printf("No end marker, skipping %s\n", url_buffer);
            continue;
        }
        pattern_end[0] = 0; // terminator for pattern

        pcre_regexp1 = pcre_compile(regexp_start, 0, &error, &erroffset,
                  NULL);

        if(!pcre_regexp1) {
            printf("Wrong regexp, skipping %s\n", url_buffer);
            printf("Error: %s at %d\n", error, erroffset);
            continue;
        }
        pcre_regexp_arr[regexp_count] = pcre_regexp1;

        len = (unsigned int)pattern_end - (unsigned int)pattern_start + 1;
        if(pattern_buff_sz+(int)len > pattern_buff_max_sz) {
            // reallocate pattern storage
            pattern_buff_max_sz *= 2;
            pattern_buff_tmp = (char*)malloc(pattern_buff_max_sz);
            if(!pattern_buff_tmp) {
                printf("Can't reallocate memory\n");
                return 0;
            }
            memcpy(pattern_buff_tmp, pattern_buff, pattern_buff_sz);
            free(pattern_buff);
            pattern_buff = pattern_buff_tmp;
        }
        memcpy(pattern_buff+pattern_buff_sz, pattern_start, len);
        pattern_arr[regexp_count] = pattern_buff+pattern_buff_sz;
        pattern_buff_sz += len;

        regexp_count++;
    }
    fclose(fh);

    fflush(stdout);

    ilen = 0;
    url_buffer[0] = 0;
    while(a = getch()) {
        if(!a || a==-1) {
            return 0;
        }
        if(a == '\n' || a == '\r') {
            if(!ilen) {
                continue;
            }
        } else {
            url_buffer[ilen] = a;
            ilen++;
            if(ilen < MAX_URL_LENGTH-1) {
                continue;
            }
        }
        url_buffer[ilen] = 0;
        ilen = 0;
        pos = strchr((const char*)url_buffer, ' ');
        if(!pos) {
            len = strlen(url_buffer);
        } else {
            len = (unsigned int)pos - (unsigned int)url_buffer;
            url_buffer[len] = 0;
        }

        for(i=0; i<regexp_count; i++) {

            rc = pcre_exec(
              pcre_regexp_arr[i],   /* the compiled pattern */
              NULL,                 /* no extra data - we didn't study the pattern */
              url_buffer,           /* the subject string */
              len,                  /* the length of the subject */
              0,                    /* start at offset 0 in the subject */
              0,                    /* default options */
              ovector,              /* output vector for substring information */
              ovector_max_sz);      /* number of elements in the output vector */

            /* Matching failed: handle error cases */
            if (rc < 0) {
                continue;
            }
            /* Match succeded */
            if (rc == 0) {
                rc = ovector_max_sz/3;
                //printf("ovector only has room for %d captured substrings\n", rc - 1);
                continue;
            }
            pattern_start = pattern_arr[i];
            url2_buffer_ptr = url2_buffer;

            while(1) {
                pos = strchr(pattern_start, '$');
                if(!pos) {
                    strcpy(url2_buffer_ptr, pattern_start);
                    break;
                }
                plen = (unsigned int)pos - (unsigned int)pattern_start;
                memcpy(url2_buffer_ptr, pattern_start, plen);
                url2_buffer_ptr += plen;
                pos++;
                j=0;
                while((a=pos[0]) >= '0' && a<='9') {
                    j = j*10+(a-'0');
                    pos++;
                }
                if(j<=rc) {
                    char *substring_start = url_buffer + ovector[2*j];
                    int substring_length = ovector[2*j+1] - ovector[2*j];
                    memcpy(url2_buffer_ptr,
                           substring_start, 
                           substring_length);
                    url2_buffer_ptr += substring_length;
                }
                pattern_start = pos;
            }
            url2_buffer_ptr[0] = '\n';
            plen = (unsigned int)url2_buffer_ptr - (unsigned int)url2_buffer;
            fwrite(url2_buffer, plen+1, 1, stdout);
            fflush(stdout);
            do_log(url2_buffer, plen+1);
            break;

        }

        if(i>=regexp_count) {
            url_buffer[len] = '\n';
            fwrite(url_buffer, len+1, 1, stdout);
            fflush(stdout);
            do_log(url_buffer, len+1);
        }
    }

    return 0;
}
